# MIT License
#
# Copyright (c) 2022 jvherck (https://jvherck.github.io/dicebear/)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import requests as r
from urllib.parse import quote
from string import ascii_lowercase, digits

import os
import pathlib
import io
from ast import literal_eval

from .errors import *
from .models import *
from .models import _FindPil

try:
    from PIL import Image as i
except Exception:
    class i:
        class Image:
            def show(self):
                pass
    _FindPil.found = False


_url = "https://avatars.dicebear.com/api/{}/{}.png?"


class DAvatar:
    """
    Base class for the avatar generator.
    """
    default_options: dict = default_options
    all_options: list = options
    def __init__(self, style: DStyle = None, seed: str = None, *, options: DOptions = None, specific_options: dict = None) -> None:
        """
        Create an avatar using this class, use `.url_svg` to get the svg url or `.url_png` to get the png url.
        Clickable links: https://github.com/jvherck/dicebear#styles , https://github.com/jvherck/dicebear#base-options , https://github.com/jvherck/dicebear#specific-style-options

        :param style: class `dicebear.models.DStyle` :: the style of avatar you want to create; check the whole list at https://github.com/jvherck/dicebear#styles
        :type style: dicebear.models.DStyle
        :param seed: class `str` :: the seed for the avatar; the avatar will be edited according to the seed
        :type seed: str
        :param options: class `DOptions` :: the options for the avatar; check the whole list at https://github.com/jvherck/dicebear#base-options
        :type options: dicebear.models.DOptions
        :param specific_options: `class: dict` :: specific options for the specified avatar style; see all specific options at https://github.com/jvherck/dicebear#specific-style-options
        :type specific_options: dict
        """
        if style is None:
            style = DStyle.random()
        if style not in styles:
            raise Error("Invalid Style", '"{}" is not a valid style! Use `DStyle.list` to see all available styles'.format(style))
        if seed is None:
            seed = "".join(choices(ascii_lowercase + digits, k=20))
        if options is None:
            options = DOptions.empty
        if specific_options is None:
            specific_options = {}

        self.__style: DStyle = style
        self.__seed: str = seed
        self.__options: DOptions = options
        self.__specific: dict = specific_options
        self.__url_svg: str = None
        self.__text: str = None
        self.__url_png: str = None
        self.__content: bytes = None
        self.__bytes: io.BytesIO = None
        self.__get_avatar_url()

    @property
    def style(self) -> DStyle:
        """
        :return: the style of the avatar
        """
        return self.__style
    @property
    def seed(self) -> str:
        """
        :return: the seed of the avatar
        """
        return self.__seed
    @property
    def options(self) -> DOptions:
        """
        :return: the options of the avatar
        """
        _option_list = {}
        for key in self.__options:
            if self.__options[key] != DAvatar.default_options[key]:
                _option_list.update({key: self.__options[key]})
        return DOptions(fromdict=_option_list)
    @property
    def url_svg(self) -> str:
        """
        :return: url to svg avatar
        """
        return self.__url_svg
    @property
    def url_png(self) -> str:
        """
        :return: url to png avatar
        """
        return self.__url_png
    @property
    def text(self) -> str:
        """
        :return: returns the bytes of the avatar in str format
        """
        return self.__text
    @property
    def bytes(self) -> bytes:
        """
        :return: returns the bytes of the avatar in bytes format
        """
        return self.__content

    def __repr__(self):
        # self.__get_avatar_url()
        return self.text

    def __str__(self) -> str:
        # self.__get_avatar_url()
        return self.__url_png

    def __eq__(self, other):
        return self.__url_png == other.__url_png
    def __ne__(self, other):
        return self.__url_png != other.__url_png
    def __le__(self, other):
        return self.bytes <= other.bytes
    def __lt__(self, other):
        return self.bytes < other.bytes
    def __ge__(self, other):
        return self.bytes >= other.bytes
    def __gt__(self, other):
        return self.bytes > other.bytes


    def __get_avatar_url(self) -> None:
        _link = _url
        _options = []
        for item in self.__options:
            if item not in all_options:
                continue
            else:
                _options.append(
                    "{}={}".format(
                        quote(item),
                        str(self.__options[item]).lower() if type(self.__options[item]) == bool else quote(str(self.__options[item]))
                    )
                )
        _specoptions = []
        for item in self.__specific:
            _specoptions.append(
                "{}={}".format(
                    quote(item),
                    str(self.__specific[item]).lower() if type(self.__specific[item]) == bool else quote(
                        str(self.__specific[item]))
                )
            )
        _link += "&".join(_options)
        if len(_options) == 0 and len(_specoptions) > 0:
            _link += "&".join(_specoptions)
        else:
            _link += "&" + "&".join(_specoptions)
        _link.replace("False", "false").replace("True", "true")
        _link = _link.format(quote(str(self.__style)), quote(self.__seed) if self.__seed is not None else "")
        req = r.get(_link)
        status = ""
        try:
            status = literal_eval(req.text)
        except ValueError:
            pass
        if type(status) == dict and "statusCode" in status:
            raise HTTPError(status)

        self.__url_png = req.url
        self.__url_svg = self._to_svg()
        self.__text = req.text
        self.__content = req.content
        # self.__response: r.Response = req
        self.__bytes = io.BytesIO(req.content)

    @staticmethod
    def __uniquify(path) -> str:
        filename, extension = os.path.splitext(path)
        counter = 1
        while os.path.exists(path):
            path = filename + "(" + str(counter) + ")" + extension
            counter += 1
        return path


    def edit(self, *, style: DStyle = None, seed: str = None, extra_options: DOptions = None, blank_options: DOptions = None) -> str:
        """
        Edit an already existing avatar.

        :param style: class `DStyle` :: edit the avatar's style (style of drawing)
        :type style: dicebear.models.DStyle
        :param seed: class `str` :: edit the avatar's seed (string to determine its looks)
        :type seed: str
        :param extra_options: class `DOptions` :: edit the avatar's options (old options stay, these get added) -- cannot be used at the same time with `blank_options` !
        :type extra_options: dicebear.models.DOptions
        :param blank_options: class `DOptions` :: reset old options and set these options as new ones (new options) -- cannot be used at the same with `extra_options` !
        :type blank_options: dicebear.models.DOptions
        :return: class `str` :: returns the link to the avatar url (png)
        """
        if style: self.__style = style
        if seed: self.__seed = seed
        if extra_options: self.__options.update(extra_options)
        elif blank_options: self.__options = blank_options
        self.__get_avatar_url()
        return self.__url_png

    def edit_specific(self, *, extra_options: dict = None, blank_options: dict = None) -> str:
        """
        Edit the specific options for an already existing avatar.

        :param extra_options: class `DOptions` :: edit the avatar's specific options (old options stay, these get added) -- cannot be used at the same time with `blank_options` !
        :type extra_options: dicebear.models.DOptions
        :param blank_options: class `DOptions` :: reset old specific options and set these options as new ones (new options) -- cannot be used at the same with `extra_options` !
        :type blank_options: dicebear.models.DOptions
        :return: class `str` :: returns the link to the avatar url (png)
        """
        if extra_options: self.__specific.update(extra_options)
        elif blank_options: self.__specific = blank_options
        self.__get_avatar_url()
        return self.__url_png

    def _to_png(self) -> str:
        """
        Converts to png and returns the url.

        :return: class `str` :: link to png avatar
        """
        self.__url_png = self.__url_svg.replace(".svg", ".png")
        return self.__url_png

    def _to_svg(self) -> str:
        """
        Converts to svg and returns the url.

        :return: class `str` :: link to png avatar
        """
        self.__url_svg = self.__url_png.replace(".png", ".svg")
        return self.__url_svg

    def save(self, *,
             location: pathlib.Path = None,
             file_name: str = "dicebear_avatar",
             file_format: DFormat = DFormat.png,
             overwrite: bool = False,
             open_after_save: bool = False
             ) -> str:
        """
        Save the avatar to your device.

        :param location: class `pathlib.Path` :: the folder to save the file in. (default is None which saves it in the current directory `os.getcwd()`
        :type location: pathlib.Path
        :param file_name: class `str` :: the name of the file to save. (default is "dicebear_avatar")
        :type file_name: str
        :param file_format: class `DFormat` :: the format of the file. (default is "png")
        :type file_format: dicebear.models.DFormat
        :param overwrite: class `bool` ::  whether to overwrite an already existing file if it has the same file name and extension
        :type overwrite: bool
        :param open_after_save: class `bool` ::  whether to open the file after saving it to your device
        :type open_after_save: bool
        :return: class `str` :: the path when successful
        """
        if file_format not in DFormat.all_formats:
            s = f'"{file_format}" is not a supported format!'
            raise ImageError(s)
        if location is None:
            location = pathlib.Path(os.getcwd())
        _location = os.path.join(location, "{}.{}".format(file_name, file_format))
        _location = self.__uniquify(_location) if overwrite is False else _location
        if file_format == DFormat.svg:
            svg_text = r.request('GET', self.__url_svg).text
        else:
            img = io.BytesIO(self.__content)
            # img = Image.open(io.BytesIO(self.__response.content))
        ret = -1
        try:
            if file_format == DFormat.svg:
                with open(_location, "w", encoding="UTF-8") as f:
                    f.write(svg_text)
                f.close()
            else:
                with open(_location, "wb") as f:
                    f.write(img.read())
                f.close()
                # img.save(_location, format)
        except ValueError:
            raise ImageValueError()
        except OSError:
            raise ImageOSError()
        except Exception as e:
            raise e
        else:
            ret = _location

        if open_after_save:
            self.view()
        return ret


    @pilcheck
    def pillow(self) -> i.Image:
        """
        Convert a :py:class:`DAvatar` to a :py:class:`PIL.Image.Image` object.

        :return: :py:class:`PIL.Image.Image`
        :raise `dicebear.errors.PILError`:
        """
        raw_img = i.open(self.__bytes).tobytes()
        img: i.Image = i.frombytes("RGBA", (256, 256), raw_img)
        return img


    def view(self, *, use_pil: bool = True) -> None:
        """
        Open and view the avatar on your device.

        :param use_pil: class `bool` :: whether to use Pillow module to open the avatar image
        :type use_pil: bool
        :return: :py:class:`NoneType`
        :raise `dicebear.errors.PILError`:
        """
        if use_pil and _FindPil.found is True:
            self.__view_pil()
        else:
            os.startfile(self.url_png, "open")

    open = view

    @pilcheck
    def __view_pil(self) -> None:
        self.pillow().show()



# MIT License
#
# Copyright (c) 2022 jvherck (https://jvherck.github.io/dicebear/)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
import traceback

from .errors import *
from random import choice, choices

class _FindPil:
    """Not important for you, just makes things easier for me on the back-end ;)"""
    found: bool = True

_ascii_lowercase = "abcdef"
_incorrect_lowercase = "ghijklmnopqrstuvwxyz"
_digits = "0123456789"

options = all_options = ["dataUri", "flip", "rotate", "scale", "radius", "size", "backgroundColor", "translateX",
                         "translateY"]
styles = ["adventurer", "adventurer-neutral", "avataaars", "big-ears", "big-ears-neutral",
          "big-smile", "bottts", "croodles", "croodles-neutral", "identicon", "initials",
          "micah", "miniavs", "open-peeps", "personas", "pixel-art", "pixel-art-neutral"]
styles_depricated = ["female", "gridy", "human", "jdenticon", "male"]


class DColor:
    """
    Base class for DAvatar's colors.
    """

    def __init__(self, html_code: str = "#ffffff"):
        """
        Colors used in this package. This uses HTML color codes!

        :param html_code: class `str` :: the html color code to use as color (default: #ffffff)
        :type html_code: str
        :raise dicebear.errors.IncorrectColor:
        """
        if html_code in ["random", "rnd"]:
            html_code = '#' + ''.join(choices(_ascii_lowercase + _digits, k=6))
        if not html_code.startswith("#"):
            html_code = "#" + html_code
        if len(html_code) not in [4, 7] or any(x in html_code for x in _incorrect_lowercase):
            raise IncorrectColor(str(html_code))
        self.html_code: str = str(html_code)

    def __str__(self):
        return f"{self.html_code}"

    def __repr__(self):
        return f"{self.html_code}"

    def __eq__(self, other):
        return self.html_code == other.html_code

    def __ne__(self, other):
        return self.html_code != other.html_code

    @staticmethod
    def random():
        """
        Get a random html code.

        :return: class `dicebear.models.DColor`
        """
        return DColor('#' + ''.join(choices(_ascii_lowercase + _digits, k=6)))


class DStyle:
    """
    All possible styles for the avatars. Visit https://avatars.dicebear.com/styles to see what they look like.\n
    - Note: Only works with attributes!
    """
    list = all_styles = styles
    adventurer = styles[0]
    adventurer_neutral = styles[1]
    avataaars = styles[2]
    big_ears = styles[3]
    big_ears_neutral = styles[4]
    big_smile = styles[5]
    bottts = styles[6]
    croodles = styles[7]
    croodles_neutral = styles[8]
    identicon = styles[9]
    initials = styles[10]
    micah = styles[11]
    miniavs = styles[12]
    open_peeps = styles[13]
    personas = styles[14]
    pixel_art = styles[15]
    pixel_art_neutral = styles[16]

    def __init__(self):
        """Only use `.attribute` to use a style."""
        pass

    @staticmethod
    def random():
        """
        Get a random style.
        """
        return choice(styles)

    @staticmethod
    def from_str(style_str: str):
        """
        Get an avatar style from a string.

        :param style_str: class `str` :: the string to convert to a DStyle
        :type style_str: str
        """
        return eval("DStyle.{}".format(style_str.replace("-", "_")))


class DFormat:
    """
    All possible image formats for saving or converting avatars.
    - Note: Only works with attributes!
    """
    list = all_formats = ["png", "svg"]
    png = "png"
    svg = "svg"

    def __init__(self):
        """Only use `.attribute` to use a format."""
        pass

    @staticmethod
    def from_str(format_str: str):
        """
        Get an avatar format from a string

        :param format_str: class `str` :: the string to convert to a DFormat
        :type format_str: str
        """
        return eval("DFormat.{}".format(format_str))


default_options: dict = {options[0]: False, options[1]: False, options[2]: 0, options[3]: 100, options[4]: 0,
                         options[5]: 0, options[6]: DColor(), options[7]: 0, options[8]: 0}


class DOptions(dict):
    """
    The options class for :py:class:`dicebear.avatar.DAvatar`
    """
    empty: dict = {}
    default_options = default = default_options

    def __init__(self, *, dataUri: bool = False, flip: bool = False, rotate: int = 0, scale: int = 100,
                 radius: int = 0, size: int = 0, backgroundColor: DColor = DColor(), translateX: int = 0,
                 translateY: int = 0, **kwargs):
        """
        Go to https://github.com/jvherck/dicebear#base-options to see all info (important for minimum and maximum values for each option!)

        :param kwargs: `fromdict` to use a custom dict instead of args (if you use this kwarg all other args will be neglected)
        """
        # kwargs: list = [dataUri, flip, rotate, scale, radius, size, backgroundColor, translateX, translateY]
        dic = kwargs.get("fromdict", {})
        current: dict = {"dataUri": dataUri, "flip": flip, "rotate": rotate, "scale": scale, "radius": radius,
                         "size": size,
                         "backgroundColor": backgroundColor, "translateX": translateX, "translateY": translateY}
        if dic:
            for item in dic:
                if item not in default_options.keys() or dic[item] == default_options[item]:
                    dic.pop(item, None)
        else:  # if not dic
            for item in current:
                if item in default_options and current[item] != default_options[item]:
                    dic.update({item: current[item]})
        if "size" in dic and dic["size"] == 0:
            dic.pop("size")
        super().__init__(dic)

    @property
    def _ver(self):
        return type(self)


def pilcheck(func):
    """Decorator to check if package Pillow is installed"""

    def wrapper(*args, **kwargs):
        if _FindPil.found is True:
            return func(*args, **kwargs)
        else:
            log_error(PILError())

    return wrapper



# MIT License
#
# Copyright (c) 2022 jvherck (https://jvherck.github.io/dicebear/)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Union
import logging

class IncorrectColor(Exception):
    """Incorrect color"""
    def __init__(self, wrong_color: str = None):
        super().__init__('Incorrect color given: "{}" is not an html hex code! (example: #ffffff)'.format(wrong_color))


class InvalidOption(Exception):
    """Invalid option"""
    def __init__(self, wrong_option: str = None):
        super().__init__('Invalid option given: "{}" is not an existing option! (use `Avatar.options` to get all possible options)'.format(
            wrong_option))



class Error(Exception):
    """General error"""
    def __init__(self, error_type: str = "", message: str = ""):
        super().__init__('{}{}'.format(error_type + (': ' if error_type else ''), message))


class HTTPError(Error):
    """HTTP error"""
    def __init__(self, dic: dict):
        super().__init__(message=str(dic))



class ImageError(Exception):
    """General image error"""
    def __init__(self, message: str = None):
        super().__init__(message)


class ImageValueError(ImageError):
    """Image value error"""
    def __init__(self, file_name: str = None):
        super().__init__('The output format could not be determined from the file name ("{}")'.format(file_name))


class ImageOSError(ImageError):
    """Image OS error"""
    def __init__(self, message: str = None):
        super().__init__('The file could not be written. The file may have been created, and may contain partial data. ("{}")'.format(message))


class PILError(ImageError):
    """Pillow error"""
    def __init__(self, message: str = "To use this function you need to install Pillow."):
        super().__init__('Module "PIL (=Pillow)" is not found! {}'.format(message))



_error_logger = logging.getLogger()
_error_handler = logging.StreamHandler()
_error_logger.setLevel(logging.ERROR)
_error_handler.setLevel(logging.ERROR)
_error_logger.addHandler(_error_handler)

def log_error(exception: Union[Exception, str], raise_error: bool = False) -> None:
    """
    Log an error.

    :param exception: class `Exception` | `str` :: the exception or string to log as error
    :type exception: Union[Exception, str]
    :param raise_error: class `bool` :: whether to raise this exception or not (default: False)
    :type raise_error: bool
    :return:
    """
    if raise_error is False:
        _error_handler.setFormatter(logging.Formatter(f"%(levelname)s: {exception.__class__.__name__}: %(message)s"))
        _error_logger.error(exception)
        return
    raise Exception(exception) if type(exception) == str else exception




# MIT License
#
# Copyright (c) 2022 jvherck (https://jvherck.github.io/dicebear/)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from setuptools import setup, find_packages
from dicebear import __version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

DESCRIPTION = "A python API wrapper for DiceBear's avatar generating API."

setup(
    name="dicebear",
    version=__version__,
    license="MIT License",
    author="jvherck",
    author_email="<jvh.discord@gmail.com>",
    maintainer="jvherck",
    maintainer_email="<jvh.discord@gmail.com>",
    url="https://github.com/jvherck/dicebear",
    project_urls={"Documentation": "https://jvherck.github.io/dicebear", "Source": "https://github.com/jvherck/dicebear", "Tracker": "https://github.com/jvherck/dicebear/issues"},
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=['requests', 'urllib3'],
    python_requires='>=3.7',
    keywords=['python', 'dicebear', 'avatar', 'avatars', 'generating', 'generation', 'generator', 'API', 'wrapper', 'image', 'images', 'picture', 'pictures', 'png', 'svg', 'cli',
              'adventurer', 'adventurer-neutral', 'avataaars', 'big-ears', 'big-ears-neutral', 'big-smile', 'bottts', 'croodles', 'croodles-neutral', 'identicon', 'initials',
              'micah', 'miniavs', 'open-peeps', 'personas', 'pixel-art', 'pixel-art-neutral'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: Unix",
        "Operating System :: MacOS", # :: MacOS X
        "Operating System :: Microsoft :: Windows",
        "License :: OSI Approved :: MIT License",
        "Topic :: Artistic Software",
        "Topic :: Games/Entertainment",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Other/Nonlisted Topic",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)
